# This code has been taken from PyTube
# https://github.com/nficano/pytube
# and is based on various patches required to keep PyTube usable.

# We need to do this since PyTube isn't being maintained.

# The below code is LICENSED under:
"""
Copyright (c) 2019 Nick Ficano
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sub-license, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:
The above copyright notice, and every other copyright notice found in this
software, and all the attributions in every file, and this permission notice
shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
# The original LICENSE notice can be found on
# https://github.com/nficano/pytube/blob/master/LICENSE


import pytube
from pytube import request
from pytube.extract import apply_signature, js_url
from typing import Any

def apply_patches():
    """
    This methods applies all the needed patches on PyTube.
    """
    # Patch 1: https://github.com/nficano/pytube/pull/643
    pytube.__main__.apply_descrambler = apply_descrambler
    # Patch 2: https://github.com/nficano/pytube/pull/634
    pytube.__main__.YouTube.extract_title = extract_title
    pytube.__main__.YouTube.descramble = descramble
    # Patch 3: https://github.com/nficano/pytube/pull/701
    pytube.cipher.get_initial_function_name = get_initial_function_name
    # Patch 4: https://github.com/nficano/pytube/pull/726
    pytube.__main__.get_ytplayer_config = get_ytplayer_config


# The below imports are required by the patches
import re
import json
from html import unescape
from urllib.parse import parse_qs, parse_qsl, unquote

def apply_descrambler(stream_data, key):
    """Apply various in-place transforms to YouTube's media stream data.
    Creates a ``list`` of dictionaries by string splitting on commas, then
    taking each list item, parsing it as a query string, converting it to a
    ``dict`` and unquoting the value.
    :param dict stream_data:
        Dictionary containing query string encoded values.
    :param str key:
        Name of the key in dictionary.
    **Example**:
    >>> d = {'foo': 'bar=1&var=test,em=5&t=url%20encoded'}
    >>> apply_descrambler(d, 'foo')
    >>> print(d)
    {'foo': [{'bar': '1', 'var': 'test'}, {'em': '5', 't': 'url encoded'}]}
    """
    otf_type = "FORMAT_STREAM_TYPE_OTF"

    if key == "url_encoded_fmt_stream_map" and not stream_data.get(
        "url_encoded_fmt_stream_map"
    ):
        formats = []
        streamingData = json.loads(stream_data["player_response"])["streamingData"]
        if 'formats' in streamingData.keys():
            formats.extend(json.loads(stream_data["player_response"])["streamingData"]["formats"])
        if 'adaptiveFormats' in streamingData.keys():
            formats.extend(json.loads(stream_data["player_response"])["streamingData"]["adaptiveFormats"])
        try:
            stream_data[key] = [
                {
                    "url": format_item["url"],
                    "type": format_item["mimeType"],
                    "quality": format_item["quality"],
                    "itag": format_item["itag"],
                    "bitrate": format_item.get("bitrate"),
                    "is_otf": (format_item.get("type") == otf_type),
                }
                for format_item in formats
            ]
        except KeyError:
            cipher_url = []
            for data in formats:
                cipher = data.get("cipher") or data["signatureCipher"]
                cipher_url.append(parse_qs(cipher))
            stream_data[key] = [
                {
                    "url": cipher_url[i]["url"][0],
                    "s": cipher_url[i]["s"][0],
                    "type": format_item["mimeType"],
                    "quality": format_item["quality"],
                    "itag": format_item["itag"],
                    "bitrate": format_item.get("bitrate"),
                    "is_otf": (format_item.get("type") == otf_type),
                }
                for i, format_item in enumerate(formats)
            ]
    else:
        stream_data[key] = [
            {k: unquote(v) for k, v in parse_qsl(i)}
            for i in stream_data[key].split(",")
        ]


def extract_title(self):
    html_lower = self.watch_html.lower()
    i_start = html_lower.index('<meta property="og:title" content="') + len(
        '<meta property="og:title" content="'
    )
    curr_i = i_start
    end_found = False
    while not end_found:
        # search for the end of the tag: ">
        if html_lower[curr_i] == '"' and html_lower[curr_i + 1] == ">":
            i_end = curr_i
            end_found = True
        curr_i += 1

    return self.watch_html[i_start:i_end].strip()


def descramble(self) -> None:
    """Descramble the stream data and build Stream instances.
    The initialization process takes advantage of Python's
    "call-by-reference evaluation," which allows dictionary transforms to
    be applied in-place, instead of holding references to mutations at each
    interstitial step.
    :rtype: None
    """
    self.vid_info = dict(parse_qsl(self.vid_info_raw))
    if self.age_restricted:
        self.player_config_args = self.vid_info
    else:
        assert self.watch_html is not None
        self.player_config_args = get_ytplayer_config(self.watch_html)["args"]

        # Fix for KeyError: 'title' issue #434
        if "title" not in self.player_config_args:  # type: ignore
            title = self.extract_title()
            self.player_config_args["title"] = unescape(title)

    # https://github.com/nficano/pytube/issues/165
    stream_maps = ["url_encoded_fmt_stream_map"]
    if "adaptive_fmts" in self.player_config_args:
        stream_maps.append("adaptive_fmts")

    # unscramble the progressive and adaptive stream manifests.
    for fmt in stream_maps:
        if not self.age_restricted and fmt in self.vid_info:
            apply_descrambler(self.vid_info, fmt)
        apply_descrambler(self.player_config_args, fmt)

        if not self.js:
            if not self.embed_html:
                self.embed_html = request.get(url=self.embed_url)
            self.js_url = js_url(self.embed_html)
            self.js = request.get(self.js_url)

        apply_signature(self.player_config_args, fmt, self.js)

        # build instances of :class:`Stream <Stream>`
        self.initialize_stream_objects(fmt)

    # load the player_response object (contains subtitle information)
    self.player_response = json.loads(self.player_config_args["player_response"])
    del self.player_config_args["player_response"]
    self.stream_monostate.title = self.title
    self.stream_monostate.duration = self.length


def get_initial_function_name(js: str) -> str:
    """Extract the name of the function responsible for computing the signature.
    :param str js:
        The contents of the base.js asset file.
    :rtype: str
    :returns:
       Function name from regex match
    """

    function_patterns = [
        r"\b[cs]\s*&&\s*[adf]\.set\([^,]+\s*,\s*encodeURIComponent\s*\(\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
        r"\b[a-zA-Z0-9]+\s*&&\s*[a-zA-Z0-9]+\.set\([^,]+\s*,\s*encodeURIComponent\s*\(\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
        r'(?:\b|[^a-zA-Z0-9$])(?P<sig>[a-zA-Z0-9$]{2})\s*=\s*function\(\s*a\s*\)\s*{\s*a\s*=\s*a\.split\(\s*""\s*\)', # noqa: E501
        r'(?P<sig>[a-zA-Z0-9$]+)\s*=\s*function\(\s*a\s*\)\s*{\s*a\s*=\s*a\.split\(\s*""\s*\)',  # noqa: E501
        r'(["\'])signature\1\s*,\s*(?P<sig>[a-zA-Z0-9$]+)\(',
        r"\.sig\|\|(?P<sig>[a-zA-Z0-9$]+)\(",
        r"yt\.akamaized\.net/\)\s*\|\|\s*.*?\s*[cs]\s*&&\s*[adf]\.set\([^,]+\s*,\s*(?:encodeURIComponent\s*\()?\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
        r"\b[cs]\s*&&\s*[adf]\.set\([^,]+\s*,\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
        r"\b[a-zA-Z0-9]+\s*&&\s*[a-zA-Z0-9]+\.set\([^,]+\s*,\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
        r"\bc\s*&&\s*a\.set\([^,]+\s*,\s*\([^)]*\)\s*\(\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
        r"\bc\s*&&\s*[a-zA-Z0-9]+\.set\([^,]+\s*,\s*\([^)]*\)\s*\(\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
        r"\bc\s*&&\s*[a-zA-Z0-9]+\.set\([^,]+\s*,\s*\([^)]*\)\s*\(\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
    ]
    for pattern in function_patterns:
        regex = re.compile(pattern)
        function_match = regex.search(js)
        if function_match:
            return function_match.group(1)

    raise RegexMatchError(
        caller="get_initial_function_name", pattern="multiple"
    )

def get_ytplayer_config(html: str) -> Any:
    """Get the YouTube player configuration data from the watch html.
    Extract the ``ytplayer_config``, which is json data embedded within the
    watch html and serves as the primary source of obtaining the stream
    manifest data.
    :param str html:
        The html contents of the watch page.
    :rtype: str
    :returns:
        Substring of the html containing the encoded manifest data.
    """
    config_patterns = [
        r";ytplayer\.config\s*=\s*({.+?});ytplayer",
        r";yt\.setConfig\(\{'PLAYER_CONFIG':\s*({.*})}\);",
        r";yt\.setConfig\(\{'PLAYER_CONFIG':\s*({.*})(,'EXPERIMENT_FLAGS'|;)",  # noqa: E501
    ]
    for pattern in config_patterns:
        regex = re.compile(pattern)
        function_match = regex.search(html)
        if function_match:
            yt_player_config = function_match.group(1)
            return json.loads(yt_player_config)

    raise RegexMatchError(
        caller="get_ytplayer_config", pattern="config_patterns"
    )

apply_patches()

YouTube = pytube.YouTube


#! Basic necessities to get the CLI running
from spotdl.search.spotifyClient import initialize
from sys import argv as cliArgs

#! Song Search from different start points
from spotdl.search.utils import get_playlist_tracks, get_album_tracks, search_for_song
from spotdl.search.songObj import SongObj

#! The actual download stuff
from spotdl.download.downloader import DownloadManager

#! to avoid packaging errors
from multiprocessing import freeze_support


#! Usage is simple - call 'python __main__.py <links, search terms, tracking files seperated by spaces>
#! Eg.
#!      python __main__.py https://open.spotify.com/playlist/37i9dQZF1DWXhcuQw7KIeM?si=xubKHEBESM27RqGkqoXzgQ 'old gods of asgard Control' https://open.spotify.com/album/2YMWspDGtbDgYULXvVQFM6?si=gF5dOQm8QUSo-NdZVsFjAQ https://open.spotify.com/track/08mG3Y1vljYA6bvDt4Wqkj?si=SxezdxmlTx-CaVoucHmrUA
#!
#! Well, yeah its a pretty long example but, in theory, it should work like a charm. 
#!
#! A '.spotdlTrackingFile' is automatically  created with the name of the first song in the playlist/album or
#! the name of the song supplied. We don't really re re re-query YTM and SPotify as all relevant details are
#! stored to disk.
#!
#! Files are cleaned up on download failure.
#!
#! All songs are normalized to standard base volume. the soft ones are made louder, the loud ones, softer.
#!
#! The progress bar is synched across multiple-processes (4 processes as of now), getting the progress bar to
#! synch was an absolute pain, each process knows how much 'it' progressed, but the display has to be for the
#! overall progress so, yeah... that took time.
#!
#! spotdl will show you its true speed on longer download's - so make sure you try downloading a playlist.
#!
#! still yet to try and package this but, in theory, there should be no errors.
#!
#!                                                          - cheerio! (Michael)
#!
#! P.S. Tell me what you think. Up to your expectations?

#! Script Help
help_notice = '''
To download a song run,
    spotdl $trackUrl
    eg. spotdl https://open.spotify.com/track/08mG3Y1vljYA6bvDt4Wqkj?si=SxezdxmlTx-CaVoucHmrUA

To download a album run,
    spotdl $albumUrl
    eg. spotdl https://open.spotify.com/album/2YMWspDGtbDgYULXvVQFM6?si=gF5dOQm8QUSo-NdZVsFjAQ

To download a playlist run,
    spotdl $playlistUrl
    eg. spotdl https://open.spotify.com/playlist/37i9dQZF1DWXhcuQw7KIeM?si=xubKHEBESM27RqGkqoXzgQ

To search for and download a song (not very accurate) run,
    spotdl $songQuery
    eg. spotdl 'The HU - Sugaan Essenna'

To resume a failed/incomplete download run,
    spotdl $pathToTrackingFile
    eg. spotdl 'Sugaan Essenna.spotdlTrackingFile'

    Note, '.spotDlTrackingFiles' are automatically created during download start, they are deleted on
    download completion

You can chain up download tasks by seperating them with spaces:
    spotdl $songQuery1 $albumUrl $songQuery2 ... (order does not matter)
    eg. spotdl 'The Hu - Sugaan Essenna' https://open.spotify.com/playlist/37i9dQZF1DWXhcuQw7KIeM?si=xubKHEBESM27RqGkqoXzgQ ...

Spotdl downloads up to 4 songs in parallel - try to download albums and playlists instead of
tracks for more speed
'''

def console_entry_point():
    '''
    This is where all the console processing magic happens.
    Its super simple, rudimentary even but, it's dead simple & it works.
    '''

    if '--help' in cliArgs or '-h' in cliArgs:
        print(help_notice)

        #! We use 'return None' as a convenient exit/break from the function
        return None

    initialize(
        clientId='4fe3fecfe5334023a1472516cc99d805',
        clientSecret='0f02b7c483c04257984695007a4a8d5c'
        )
    
    downloader = DownloadManager()

    for request in cliArgs[1:]:
        if 'open.spotify.com' in request and 'track' in request:
            print('Fetching Song...')
            song = SongObj.from_url(request)

            if song.get_youtube_link() != None:
                downloader.download_single_song(song)
            else:
                print('Skipping %s (%s) as no match could be found on youtube' % (
                    song.get_song_name(), request
                ))
        
        elif 'open.spotify.com' in request and 'album' in request:
            print('Fetching Album...')
            songObjList = get_album_tracks(request)

            downloader.download_multiple_songs(songObjList)
        
        elif 'open.spotify.com' in request and 'playlist' in request:
            print('Fetching Playlist...')
            songObjList = get_playlist_tracks(request)

            downloader.download_multiple_songs(songObjList)
        
        elif request.endswith('.spotdlTrackingFile'):
            print('Preparing to resume download...')
            downloader.resume_download_from_tracking_file(request)
        
        else:
            print('Searching for song "%s"...' % request)
            try:
                song = search_for_song(request)
                downloader.download_single_song(song)

            except Exception:
                print('No song named "%s" could be found on spotify' % request)
    
    downloader.close()

if __name__ == '__main__':
    freeze_support()

    console_entry_point()


#===============
#=== Imports ===
#===============

from spotdl.download.progressHandlers import ProgressRootProcess

from os import mkdir, remove, system as run_in_shell
from os.path import join, exists

from multiprocessing import Pool

from spotdl.patches.pyTube import YouTube

from mutagen.easyid3 import EasyID3, ID3
from mutagen.id3 import APIC as AlbumCover

from urllib.request import urlopen

#! The following are not used, they are just here for static typechecking with mypy
from typing import List

from spotdl.search.songObj import SongObj
from spotdl.download.progressHandlers import DisplayManager, DownloadTracker



#==========================
#=== Base functionality ===
#==========================

#! Technically, this should ideally be defined within the downloadManager class. But due
#! to the quirks of multiprocessing.Pool, that can't be done. Do not consider this as a
#! standalone function but rather as part of DownloadManager

def download_song(songObj: SongObj, displayManager: DisplayManager = None,
                                    downloadTracker: DownloadTracker = None) -> None:
    '''
    `songObj` `songObj` : song to be downloaded

    `AutoProxy` `displayManager` : autoproxy reference to a `DisplayManager`

    `AutoProxy` `downloadTracker`: autoproxy reference to a `DownloadTracker`

    RETURNS `~`

    Downloads, Converts, Normalizes song & embeds metadata as ID3 tags.
    '''

    #! all YouTube downloads are to .\Temp; they are then converted and put into .\ and
    #! finally followed up with ID3 metadata tags

    #! we explicitly use the os.path.join function here to ensure download is
    #! platform agnostic
    
    # Create a .\Temp folder if not present
    tempFolder = join('.', 'Temp')
    
    if not exists(tempFolder):
        mkdir(tempFolder)
    
    # build file name of converted file
    artistStr = ''

    #! we eliminate contributing artist names that are also in the song name, else we
    #! would end up with things like 'Jetta, Mastubs - I'd love to change the world
    #! (Mastubs REMIX).mp3' which is kinda an odd file name.
    for artist in songObj.get_contributing_artists():
        if artist.lower() not in songObj.get_song_name().lower():
            artistStr += artist + ', '
    
    #! the ...[:-2] is to avoid the last ', ' appended to artistStr
    convertedFileName = artistStr[:-2] + ' - ' + songObj.get_song_name()

    #! this is windows specific (disallowed chars)
    for disallowedChar in ['/', '?', '\\', '*','|', '<', '>']:
        if disallowedChar in convertedFileName:
            convertedFileName = convertedFileName.replace(disallowedChar, '')
    
    #! double quotes (") and semi-colons (:) are also disallowed characters but we would
    #! like to retain their equivalents, so they aren't removed in the prior loop
    convertedFileName = convertedFileName.replace('"', "'").replace(': ', ' - ')

    convertedFilePath = join('.', convertedFileName) + '.mp3'



    # if a song is already downloaded skip it
    if exists(convertedFilePath):
        if displayManager:
            displayManager.notify_download_skip()
        if downloadTracker:
            downloadTracker.notify_download_completion(songObj)
        
        #! None is the default return value of all functions, we just explicitly define
        #! it here as a continent way to avoid executing the rest of the function.
        return None
    


    # download Audio from YouTube
    if displayManager:
        youtubeHandler = YouTube(
            url                  = songObj.get_youtube_link(),
            on_progress_callback = displayManager.pytube_progress_hook
        )
    else:
        youtubeHandler = YouTube(songObj.get_youtube_link())
    
    trackAudioStream = youtubeHandler.streams.get_audio_only()

    #! The actual download, if there is any error, it'll be here,
    try:
        #! pyTube will save the song in .\Temp\$songName.mp4, it doesn't save as '.mp3'
        downloadedFilePath = trackAudioStream.download(
            output_path   = tempFolder,
            filename      = convertedFileName,
            skip_existing = False
        )
    except:
        #! This is equivalent to a failed download, we do nothing, the song remains on
        #! downloadTrackers download queue and all is well...
        #!
        #! None is again used as a convenient exit
        remove(join(tempFolder, convertedFileName) + '.mp4')
        return None
    


    # convert downloaded file to MP3 with normalization

    #! -af loudnorm=I=-7:LRA applies EBR 128 loudness normalization algorithm with
    #! intergrated loudness target (I) set to -17, using values lower than -15
    #! causes 'pumping' i.e. rhythmic variation in loudness that should not
    #! exist -loud parts exaggerate, soft parts left alone.
    #! 
    #! dynaudnorm applies dynamic non-linear RMS based normalization, this is what
    #! actually normalized the audio. The loudnorm filter just makes the apparent
    #! loudness constant
    #!
    #! apad=pad_dur=2 adds 2 seconds of silence toward the end of the track, this is
    #! done because the loudnorm filter clips/cuts/deletes the last 1-2 seconds on
    #! occasion especially if the song is EDM-like, so we add a few extra seconds to
    #! combat that.
    #!
    #! -acodec libmp3lame sets the encoded to 'libmp3lame' which is far better
    #! than the default 'mp3_mf', '-abr true' automatically determines and passes the
    #! audio encoding bitrate to the filters and encoder. This ensures that the
    #! sampled length of songs matches the actual length (i.e. a 5 min song won't display
    #! as 47 seconds long in your music player, yeah that was an issue earlier.)

    command = 'ffmpeg -v quiet -y -i "%s" -acodec libmp3lame -abr true -af "apad=pad_dur=2, dynaudnorm, loudnorm=I=-17" "%s"'
    formattedCommand = command % (downloadedFilePath, convertedFilePath)

    run_in_shell(formattedCommand)

    #! Wait till converted file is actually created
    while True:
        if exists(convertedFilePath):
            break

    if displayManager:
        displayManager.notify_conversion_completion()



    # embed song details
    #! we save tags as both ID3 v2.3 and v2.4

    #! The simple ID3 tags
    audioFile = EasyID3(convertedFilePath)

    #! Get rid of all existing ID3 tags (if any exist)
    audioFile.delete()

    #! song name
    audioFile['title'] = songObj.get_song_name()
    audioFile['titlesort'] = songObj.get_song_name()

    #! track number
    audioFile['tracknumber'] = str(songObj.get_track_number())

    #! genres (pretty pointless if you ask me)
    #! we only apply the first available genre as ID3 v2.3 doesn't support multiple
    #! genres and ~80% of the world PC's run Windows - an OS with no ID3 v2.4 support
    genres = songObj.get_genres()

    if len(genres) > 0:
        audioFile['genre'] = genres[0]
    
    #! all involved artists
    audioFile['artist'] = songObj.get_contributing_artists()

    #! album name
    audioFile['album'] = songObj.get_album_name()

    #! album artist (all of 'em)
    audioFile['albumartist'] = songObj.get_album_artists()

    #! album release date (to what ever precision available)
    audioFile['date']         = songObj.get_album_release()
    audioFile['originaldate'] = songObj.get_album_release()

    #! save as both ID3 v2.3 & v2.4 as v2.3 isn't fully features and
    #! windows doesn't support v2.4 until later versions of Win10
    audioFile.save(v2_version = 3)

    #! setting the album art
    audioFile = ID3(convertedFilePath)

    rawAlbumArt = urlopen(songObj.get_album_cover_url()).read()

    audioFile['APIC'] = AlbumCover(
        encoding = 3,
        mime = 'image/jpeg',
        type = 3,
        desc = 'Cover',
        data = rawAlbumArt
    )

    audioFile.save(v2_version = 3)

    # Do the necessary cleanup
    if displayManager:
        displayManager.notify_download_completion()
    
    if downloadTracker:
        downloadTracker.notify_download_completion(songObj)



    # delete the unnecessary YouTube download File
    remove(downloadedFilePath)


#===========================================================
#=== The Download Manager (the tyrannical boss lady/guy) ===
#===========================================================

class DownloadManager():
    #! Big pool sizes on slow connections will lead to more incomplete downloads
    poolSize = 4

    def __init__(self):
        # start a server for objects shared across processes
        progressRoot = ProgressRootProcess()
        progressRoot.start()

        self.rootProcess = progressRoot

        # initialize shared objects
        self.displayManager  = progressRoot.DisplayManager()
        self.downloadTracker = progressRoot.DownloadTracker()

        self.displayManager.clear()

        # initialize worker pool
        self.workerPool = Pool( DownloadManager.poolSize )
    
    def download_single_song(self, songObj: SongObj) -> None:
        '''
        `songObj` `song` : song to be downloaded

        RETURNS `~`

        downloads the given song
        '''

        self.downloadTracker.clear()
        self.downloadTracker.load_song_list([songObj])

        self.displayManager.reset()
        self.displayManager.set_song_count_to(1)

        download_song(songObj, self.displayManager, self.downloadTracker)

        print()
    
    def download_multiple_songs(self, songObjList: List[SongObj]) -> None:
        '''
        `list<songObj>` `songObjList` : list of songs to be downloaded

        RETURNS `~`

        downloads the given songs in parallel
        '''

        self.downloadTracker.clear()
        self.downloadTracker.load_song_list(songObjList)

        self.displayManager.reset()
        self.displayManager.set_song_count_to(len(songObjList))

        self.workerPool.starmap(
            func     = download_song,
            iterable = (
                (song, self.displayManager, self.downloadTracker)
                    for song in songObjList
            )
        )
        print()
    
    def resume_download_from_tracking_file(self, trackingFilePath: str) -> None:
        '''
        `str` `trackingFilePath` : path to a .spotdlTrackingFile

        RETURNS `~`

        downloads songs present on the .spotdlTrackingFile in parallel
        '''

        self.downloadTracker.clear()
        self.downloadTracker.load_tracking_file(trackingFilePath)

        songObjList = self.downloadTracker.get_song_list()

        self.displayManager.reset()
        self.displayManager.set_song_count_to(len(songObjList))

        self.workerPool.starmap(
            func     = download_song,
            iterable = (
                (song, self.displayManager, self.downloadTracker)
                    for song in songObjList
            )
        )
        print()
    
    def close(self) -> None:
        '''
        RETURNS `~`

        cleans up across all processes
        '''

        self.displayManager.close()
        self.rootProcess.shutdown()

        self.workerPool.close()
        self.workerPool.join()


#===============
#=== Imports ===
#===============

#! the following are for the search provider to function
from rapidfuzz.fuzz import partial_ratio
from json import loads as convert_json_to_dict
from requests import post

#! Just for static typing
from typing import List



#================================
#=== Note to readers / Coders ===
#================================

#! YTM search (the actual POST request), courtesy of Elliot G. (@rocketinventor)
#! result parsing and song matching system by @Mikhail-Zex
#!
#! Essentially, Without Elliot, you wouldn't have a YTM search provider at all.



#=======================
#=== helper function ===
#=======================

def match_percentage(str1: str, str2: str, score_cutoff: float = 0) -> float:
    '''
    `str` `str1` : a random sentence

    `str` `str2` : another random sentence

    `float` `score_cutoff` : minimum score required to consider it a match
                             returns 0 when similarity < score_cutoff

    RETURNS `float`

    A wrapper around `rapidfuzz.fuzz.partial_ratio` to handle UTF-8 encoded
    emojis that usually cause errors
    '''

    #! this will throw an error if either string contains a UTF-8 encoded emoji 
    try:
        return partial_ratio(str1, str2, score_cutoff=score_cutoff)

    #! we build new strings that contain only alphanumerical characters and spaces
    #! and return the partial_ratio of that
    except:
        newStr1 = ''

        for eachLetter in str1:
            if eachLetter.isalnum() or eachLetter.isspace():
                newStr1 += eachLetter

        newStr2 = ''

        for eachLetter in str2:
            if eachLetter.isalnum() or eachLetter.isspace():
                newStr2 += eachLetter

        return partial_ratio(newStr1, newStr2, score_cutoff=score_cutoff)

#========================================================================
#=== Background functions/Variables (Not meant to be called directly) ===
#========================================================================

#! as of now there is no need to allow anyone to set a ytm-api key, so no-one should mess
#! with this. It's only defined out here instead of within a class so that chaning of the
#! api key if and when the change happens should be a simple job
ytmApiKey = 'AIzaSyC9XL3ZjWddXya6X74dJoCTL-WEYFDNX30'

def __query_and_simplify(searchTerm: str, apiKey: str = ytmApiKey) -> List[dict]:
    '''
    `str` `searchTerm` : the search term you would type into YTM's search bar

    RETURNS `list<dict>`

    For structure of dict, see comment at function declaration
    '''

    #! For dict structure, see end of this function (~ln 268, ln 283) and chill, this
    #! function ain't soo big, there are plenty of comments and blank lines



    # build and POST a query to YTM

    url = 'https://music.youtube.com/youtubei/v1/search?alt=json&key=' + apiKey

    headers = {
        'Referer': 'https://music.youtube.com/search'
    }

    payload = {

        'context': {
            'client': {
                'clientName': 'WEB_REMIX',
                'clientVersion': '0.1'
            }
        },

        'query': searchTerm
    }

    request = post(
        url     = url,
        headers = headers,
        json    = payload
    )

    response = convert_json_to_dict(request.text)

    #! We will hereon call generic levels of nesting as 'Blocks'. What follows is an
    #! overview of the basic nesting (you'll need it),
    #!
    #! Content blocks
    #!       Group results into 'Top result', 'Songs', 'Videos',
    #!       'playlists', 'Albums', 'User notices', etc...
    #!
    #! Result blocks (under each 'Content block')
    #!       Represents an individual song, album, video, ..., playlist
    #!       result
    #!
    #! Detail blocks (under each 'Result block')
    #!       Represents a detail of the result, these might be one of
    #!       name, duration, channel, album, etc...
    #!
    #! Link block (under each 'Result block')
    #!       Contains the video/Album/playlist/Song/Artist link of the
    #!       result as found on YouTube
    
    # Simplify and extract necessary details from senselessly nested YTM response
    #! This process is broken into numbered steps below

    #! nested-dict keys are used to (a) indicate nesting visually and (b) make the code
    #! more readable



    # 01. Break response into contentBLocks
    contentBlocks = response['contents'] \
        ['sectionListRenderer'] \
            ['contents']



    # 02. Gather all result block in the same place
    resultBlocks = []

    for cBlock in contentBlocks:
        # Ignore user-suggestion

        #! The 'itemSectionRenderer' field is for user notices (stuff like - 'showing
        #! results for xyz, search for abc instead') we have no use for them, the for
        #! loop below if throw a keyError if we don't ignore them
        if 'itemSectionRenderer' in cBlock.keys():
            continue

        for contents in cBlock['musicShelfRenderer']['contents']:
            #! apparently content Blocks without an 'overlay' field don't have linkBlocks
            #! I have no clue what they are and why there even exist
            if 'overlay' not in contents['musicResponsiveListItemRenderer']:
                continue

            result = contents['musicResponsiveListItemRenderer'] \
                ['flexColumns']
            


            # Add the linkBlock        
            
            linkBlock = contents['musicResponsiveListItemRenderer'] \
                ['overlay'] \
                    ['musicItemThumbnailOverlayRenderer'] \
                        ['content'] \
                            ['musicPlayButtonRenderer'] \
                                ['playNavigationEndpoint']
            
            #! detailsBlock is always a list, so we just append the linkBlock to it
            #! insted of carrying along all the other junk from 'musicResponsiveListItemRenderer'
            result.append(linkBlock)
        
            #! gather resultBlock
            resultBlocks.append(result)
    
    

    # 03. Gather available details in the same place

    #! We only need results that are Songs or Videos, so we filter out the rest, since
    #! Songs and Videos are supplied with different details, extracting all details from
    #! both is just carrying on redundant data, so we also have to selectively extract
    #! relevant details. What you need to know to understand how we do that here:
    #!
    #! Songs details are ALWAYS in the following order:
    #!       0 - Name
    #!       1 - Type (Song)
    #!       2 - Artist
    #!       3 - Album
    #!       4 - Duration (mm:ss)
    #!
    #! Video details are ALWAYS in the following order:
    #!       0 - Name
    #!       1 - Type (Video)
    #!       2 - Channel
    #!       3 - Viewers
    #!       4 - Duration (hh:mm:ss)
    #!
    #! We blindly gather all the details we get our hands on, then
    #! cherrypick the details we need based on  their index numbers,
    #! we do so only if their Type is 'Song' or 'Video

    simplifiedResults = []

    for result in resultBlocks:

        # Blindly gather available details
        availableDetails = []



        # Filterout dummies here itself
        #! 'musicResponsiveListItmeFlexColumnRenderer' should have more that one
        #! sub-block, if not its a dummy, why does the YTM response contain dummies?
        #! I have no clue. We skip these.
        
        #! Remember that we appended the linkBlock to result, treating that like the
        #! other constituents of a result block will lead to errors, hence the 'in
        #! result[:-1]'
        for detail in result[:-1]:
            if len(detail['musicResponsiveListItemFlexColumnRenderer']) < 2:
                continue
            
            #! if not a dummy, collect all available details
            availableDetails.append(
                detail['musicResponsiveListItemFlexColumnRenderer'] \
                    ['text'] \
                        ['runs'][0] \
                            ['text']
            )
        


        # Filterout non-Song/Video results and incomplete results here itself
        #! From what we know about detail order, note that [1] - indicate result type
        if availableDetails[1] in ['Song', 'Video'] and len(availableDetails) == 5:
            
            #! skip if result is in hours instead of minuts (no song is that long)
            if len(availableDetails[4].split(':')) != 2:
                    continue



            # grab position of result
            #! This helps for those oddball cases where 2+ results are rated equally,
            #! lower position --> better match
            resultPosition = resultBlocks.index(result)



            # grab result link
            #! this is nested as [playlistEndpoint/watchEndpoint][videoId/playlistId/...]
            #! so hardcoding the dict keys for data look up is an ardours process, since
            #! the sub-block pattern is fixed even though the key isn't, we just
            #! reference the dict keys by index
            endpointKey = list( result[-1].keys() )[1]
            resultIdKey = list( result[-1][endpointKey].keys() )[0]

            linkId = result[-1][endpointKey][resultIdKey]
            resultLink = 'https://www.youtube.com/watch?v=' + linkId



            # convert length into seconds
            minStr, secStr = availableDetails[4].split(':')

            #! handle leading zeroes (eg. 01, 09, etc...), they cause eval errors, there
            #! are a few oddball tracks that are only a few seconds long
            if minStr[0] == '0' and len(minStr) == 2:
                minStr = minStr[1]
            
            if secStr[0] == '0':
                secStr = secStr[1]
            
            time = eval(minStr)*60 + eval(secStr)



            # format relevant details
            if availableDetails[1] == 'Song':
            
                formattedDetails = {
                    'name'      : availableDetails[0],
                    'type'      : 'song',
                    'artist'    : availableDetails[2],
                    'album'     : availableDetails[3],
                    'length'    : time,
                    'link'      : resultLink,
                    'position'  : resultPosition
                }
            
                if formattedDetails not in simplifiedResults:
                    simplifiedResults.append(formattedDetails)
            
            elif availableDetails[1] == 'Video':
            
                formattedDetails = {
                    'name'      : availableDetails[0],
                    'type'      : 'video',
                    'length'    : time,
                    'link'      : resultLink,
                    'position'  : resultPosition
                }
            
                if formattedDetails not in simplifiedResults:
                    simplifiedResults.append(formattedDetails)
            
            #! Things like playlists, albums, etc... just get ignored
            


    # return the results
    return simplifiedResults



#=======================
#=== Search Provider ===
#=======================

def search_and_order_ytm_results(songName: str, songArtists: List[str],
                                        songAlbumName: str, songDuration: int) -> dict:
    '''
    `str` `songName` : name of song

    `list<str>` `songArtists` : list containing name of contributing artists

    `str` `songAlbumName` : name of song's album

    `int` `songDuration`

    RETURNS `dict`

    each entry in the result if formated as {'$YouTubeLink': $matchValue, ...}; Match value
    indicates how good a match the result is the the given parameters. THe maximum value
    that $matchValue can take is 100, the least value is unbound.
    '''

    # Build a YTM search term
    songArtistStr = ''

    for artist in songArtists:
        songArtistStr += artist + ', '
    
    #! the ...[:-2] is so that we don't get the last ', ' of searchStrArtists because we
    #! want $artist1, $artist2, ..., $artistN - $songName
    songSearchStr = songArtistStr[:-2] + ' - ' + songName



    # Query YTM
    results = __query_and_simplify(songSearchStr)



    # Assign an overall avg match value to each result
    linksWithMatchValue = {}

    for result in results:
        #! If there are no common words b/w the spotify and YouTube Music name, the song
        #! is a wrong match (Like Ruelle - Madness being matched to Ruelle - Monster, it
        #! happens without this conditional)
        
        #! most song results on youtube go by $artist - $songName, so if the spotify name
        #! has a '-', this function would return True, a common '-' is hardly a 'common
        #! word', so we get rid of it. Lower-caseing all the inputs is to get rid of the
        #! troubles that arise from pythons handling of differently cased words, i.e.
        #! 'Rhino' == 'rhino' is false though the word is same... so we lower-case both
        #! sentences and replace any hypens(-)
        lowerSongName = songName.lower()
        lowerResultName = result['name'].lower()

        sentenceAWords = lowerSongName.replace('-', ' ').split(' ')
        
        commonWord = False

        #! check for common word
        for word in sentenceAWords:
            if word != '' and word in lowerResultName:
                commonWord = True
        
        #! if there are no common words, skip result
        if not commonWord:
            continue



        # Find artist match
        #! match  = (no of artist names in result) / (no. of artist names on spotify) * 100
        artistMatchNumber = 0

        #! we use fuzzy matching because YouTube spellings might be mucked up
        if result['type'] == 'song':
            for artist in songArtists:
                if match_percentage (artist.lower(), result['artist'].lower(), 85):
                    artistMatchNumber += 1
        else:
            #! i.e if video
            for artist in songArtists:
                #! something like match_percentage('rionos', 'aiobahn, rionos Motivation
                #! (remix)' would return 100, so we're absolutely corrent in matching
                #! artists to song name.
                if match_percentage(artist.lower(), result['name'].lower(), 85):
                    artistMatchNumber += 1
        
        #! Skip if there are no artists in common, (else, results like 'Griffith Swank -
        #! Madness' will be the top match for 'Ruelle - Madness')
        if artistMatchNumber == 0:
            continue
        
        artistMatch = (artistMatchNumber / len(songArtists) ) * 100



        # Find name match
        nameMatch = round(match_percentage(result['name'], songName), ndigits = 3)



        # Find album match
        #! We assign an arbitrary value of 0 for album match in case of video results
        #! from YouTube Music
        albumMatch = 0

        if result['type'] == 'song':
            albumMatch = match_percentage(result['album'], songAlbumName)



        # Find duration match
        #! time match = 100 - (delta(duration)**2 / original duration * 100)
        #! difference in song duration (delta) is usually of the magnitude of a few
        #! seconds, we need to amplify the delta if it is to have any meaningful impact
        #! wen we calculate the avg match value
        delta = result['length'] - songDuration
        nonMatchValue = (delta ** 2) / songDuration * 100

        timeMatch = 100 - nonMatchValue



        # the results along with the avg Match
        avgMatch = (artistMatch + albumMatch + nameMatch + timeMatch) / 4

        linksWithMatchValue[result['link']] = avgMatch

    return linksWithMatchValue

def search_and_get_best_match(songName: str, songArtists: List[str],
                                        songAlbumName: str, songDuration: int) -> str:
    '''
    `str` `songName` : name of song

    `list<str>` `songArtists` : list containing name of contributing artists

    `str` `songAlbumName` : name of song's album

    `int` `songDuration` : duration of the song

    RETURNS `str` : link of the best match
    '''

    #! This is lazy coding, sorry.
    results = search_and_order_ytm_results(
        songName, songArtists,
        songAlbumName, songDuration
    )

    if len(results) == 0:
        return None

    resultItems = list(results.items())
    sortedResults = sorted(resultItems, key = lambda x: x[1], reverse=True)

    #! In theory, the first 'TUPLE' in sortedResults should have the highest match
    #! value, we send back only the link
    return sortedResults[0][0]
